#include "audiosys.h"

/* ---------------------- */
/*  Audio Render Handler  */
/* ---------------------- */

#include "sct_main.h"

static NES_AUDIO_HANDLER *nah = 0;
static NES_VOLUME_HANDLER *nvh = 0;
void NESAudioRender(Int16 *bufp, Uint buflen)
{
	while (buflen--)
	{
		NES_AUDIO_HANDLER *ph;
		Int32 accum = 0;

		for (ph = nah; ph; ph = ph->next)
		{
			if (!ph->fSkip || bufp) accum += ph->Proc();
		}

		if (bufp)
		{
			Uint32 output;
			accum += (0x8000 << 8);
			if (accum < 0)
				output = 0;
			else if (accum >= (0x10000 << 8))
				output = (0x10000 << 8) - 1;
			else
				output = accum;
			output >>= 8;
			*bufp++ = ((Int32)output) - 0x8000;
		}
	}
}

void NESVolume(Uint volume)
{
	NES_VOLUME_HANDLER *ph;
	for (ph = nvh; ph; ph = ph->next) ph->Proc(volume);
}

#include "sct_init.h"

void NESAudioHandlerInstall(NES_AUDIO_HANDLER *ph)
{
	/* Add to tail of list*/
	ph->next = 0;
	if (nah)
	{
		NES_AUDIO_HANDLER *p = nah;
		while (p->next) p = p->next;
		p->next = ph;
	}
	else
	{
		nah = ph;
	}
}
void NESVolumeHandlerInstall(NES_VOLUME_HANDLER *ph)
{
	/* Add to tail of list*/
	ph->next = 0;
	if (nvh)
	{
		NES_VOLUME_HANDLER *p = nvh;
		while (p->next) p = p->next;
		p->next = ph;
	}
	else
	{
		nvh = ph;
	}
}
void NESAudioHandlerInitialize(void)
{
	nah = 0;
	nvh = 0;
}



static Uint frequency = 44100;
void NESAudioFrequencySet(Uint freq)
{
	frequency = freq;
}
Uint NESAudioFrequencyGet(void)
{
	return frequency;
}
