#include "handler.h"
#include "nes6502.h"

/* ----------------------- */
/*  Memory Access Handler  */
/* ----------------------- */

#include "sct_main.h"

static NES_READ_HANDLER  *(nprh[0x10]) = { 0, };
static NES_WRITE_HANDLER *(npwh[0x10]) = { 0, };
#define EXTREADWRITE(p) \
static Uint __fastcall ExtRd##p##(Uint A) \
{ \
	NES_READ_HANDLER *ph = nprh[0x##p##]; \
	do \
	{ \
		if (ph->min <= A && A <= ph->max) \
		{ \
			return ph->Proc(A); \
		} \
	} while ((ph = ph->next) != 0); \
	return 0; \
} \
static void __fastcall ExtWr##p##(Uint A, Uint V) \
{ \
	NES_WRITE_HANDLER *ph = npwh[0x##p##]; \
	do \
	{ \
		if (ph->min <= A && A <= ph->max) \
		{ \
			ph->Proc(A, V); \
			return; \
		} \
	} while ((ph = ph->next) != 0); \
}
EXTREADWRITE(0)
EXTREADWRITE(1)
EXTREADWRITE(2)
EXTREADWRITE(3)
EXTREADWRITE(4)
EXTREADWRITE(5)
EXTREADWRITE(6)
EXTREADWRITE(7)
EXTREADWRITE(8)
EXTREADWRITE(9)
EXTREADWRITE(A)
EXTREADWRITE(B)
EXTREADWRITE(C)
EXTREADWRITE(D)
EXTREADWRITE(E)
EXTREADWRITE(F)
static READHANDLER ExtRdTbl[0x10] = {
	ExtRd0,ExtRd1,ExtRd2,ExtRd3,
	ExtRd4,ExtRd5,ExtRd6,ExtRd7,
	ExtRd8,ExtRd9,ExtRdA,ExtRdB,
	ExtRdC,ExtRdD,ExtRdE,ExtRdF,
};
static WRITEHANDLER ExtWrTbl[0x10] = {
	ExtWr0,ExtWr1,ExtWr2,ExtWr3,
	ExtWr4,ExtWr5,ExtWr6,ExtWr7,
	ExtWr8,ExtWr9,ExtWrA,ExtWrB,
	ExtWrC,ExtWrD,ExtWrE,ExtWrF,
};
static Uint __fastcall NullRead(Uint A)
{
	return 0;
}
static void __fastcall NullWrite(Uint A, Uint V)
{
}

#include "sct_init.h"

static void InstallPageReadHandler(NES_READ_HANDLER *ph)
{
	Uint page = (ph->min >> 12) & 0xF;
	if (nprh[page])
		NES6502ReadHandlerSet(page, ExtRdTbl[page]);
	else
		NES6502ReadHandlerSet(page, ph->Proc);
	/* Add to head of list*/
	ph->next = nprh[page];
	nprh[page] = ph;
}
static void InstallPageWriteHandler(NES_WRITE_HANDLER *ph)
{
	Uint page = (ph->min >> 12) & 0xF;
	if (npwh[page])
		NES6502WriteHandlerSet(page, ExtWrTbl[page]);
	else
		NES6502WriteHandlerSet(page, ph->Proc);
	/* Add to head of list*/
	ph->next = npwh[page];
	npwh[page] = ph;
}
void NESReadHandlerInstall(NES_READ_HANDLER *ph)
{
	for (; ph->Proc; ph++) InstallPageReadHandler(ph);
}
void NESWriteHandlerInstall(NES_WRITE_HANDLER *ph)
{
	for (; ph->Proc; ph++) InstallPageWriteHandler(ph);
}
static void NESMemoryHandlerInitialize(void)
{
	Uint i;
	for (i = 0; i < 0x10;  i++)
	{
		NES6502ReadHandlerSet(i, NullRead);
		NES6502WriteHandlerSet(i, NullWrite);
		nprh[i] = 0;
		npwh[i] = 0;
	}
}


/* --------------- */
/*  Reset Handler  */
/* --------------- */
static NES_RESET_HANDLER *(nrh[0x10]) = { 0, };
void NESReset(void)
{
	NES_RESET_HANDLER *ph;
	Uint prio;
	for (prio = 0; prio < 0x10; prio++)
		for (ph = nrh[prio]; ph; ph = ph->next) ph->Proc();
}
static void InstallPriorityResetHandler(NES_RESET_HANDLER *ph)
{
	Uint prio = ph->priority;
	if (prio > 0xF) prio = 0xF;
	/* Add to tail of list*/
	ph->next = 0;
	if (nrh[prio])
	{
		NES_RESET_HANDLER *p = nrh[prio];
		while (p->next) p = p->next;
		p->next = ph;
	}
	else
	{
		nrh[prio] = ph;
	}
}
void NESResetHandlerInstall(NES_RESET_HANDLER *ph)
{
	for (; ph->Proc; ph++) InstallPriorityResetHandler(ph);
}
static void NESResetHandlerInitialize(void)
{
	Uint prio;
	for (prio = 0; prio < 0x10; prio++) nrh[prio] = 0;
}


/* ------------------- */
/*  Terminate Handler  */
/* ------------------- */
static NES_TERMINATE_HANDLER *nth = 0;
void NESTerminate(void)
{
	NES_TERMINATE_HANDLER *ph;
	for (ph = nth; ph; ph = ph->next) ph->Proc();
}
void NESTerminateHandlerInstall(NES_TERMINATE_HANDLER *ph)
{
	/* Add to head of list*/
	ph->next = nth;
	nth = ph;
}
static void NESTerminateHandlerInitialize(void)
{
	nth = 0;
}


void NESHandlerInitialize(void)
{
	NESMemoryHandlerInitialize();
	NESResetHandlerInitialize();
	NESTerminateHandlerInitialize();
}
