#include "nes6502.h"

#include "sct_main.h"

#define USE_DIRECT_ZEROPAGE 0
#define External __inline static

#include "km6502/km6502m.h"

static Uint work6502_BP;	/* Break Point */
static struct K6502_Context work6502;

void NES6502ReadHandlerSet(Uint bank, READHANDLER rh)
{
	work6502.ReadByte[bank] = rh;
}

void NES6502WriteHandlerSet(Uint bank, WRITEHANDLER wh)
{
	work6502.WriteByte[bank] = wh;
}

void NES6502SetZeropage(void *p)
{
#if USE_FAST_ZEROPAGE
	work6502.zeropage = p;
#endif
}

void NES6502BreakPoint(Uint A)
{
	work6502_BP = A;
}

void NES6502Irq(void)
{
	work6502.iRequest |= K6502_INT;
}

void NES6502Nmi(void)
{
	work6502.iRequest |= K6502_NMI;
}

Uint NES6502Read(Uint A)
{
   return work6502.ReadByte[A >> USE_INLINEMMC](A);
}

Uint NES6502Execute(Uint total_cycles)
{
	while (work6502.clock < total_cycles)
	{
		K6502_Exec(&work6502);
		if (work6502.PC == work6502_BP)
		{
			work6502.clock = 0;
			return 1;
		}
	}
	work6502.clock -= total_cycles;
	return 0;
}

void NES6502ContextGet(NES6502_CONTEXT *p)
{
	p->PC = work6502.PC;
	p->A = work6502.A;
	p->P = work6502.P;
	p->X = work6502.X;
	p->Y = work6502.Y;
	p->S = work6502.S;
}

void NES6502ContextSet(const NES6502_CONTEXT *p)
{
	work6502.PC = p->PC;
	work6502.A = p->A;
	work6502.P = p->P;
	work6502.X = p->X;
	work6502.Y = p->Y;
	work6502.S = p->S;
}

#include "sct_init.h"

void NES6502Reset(void)
{
	work6502.clock = 0;
	work6502.iRequest = K6502_INIT;
	work6502.PC = work6502_BP = 0xFFFF;
	NES6502Execute(work6502.clock + 1);
}

void NES6502Initialize(void)
{
#if USE_FAST_ZEROPAGE
	work6502.zeropage = 0;
#endif
	NES6502Reset();
}
