#include "neserr.h"
#include "handler.h"
#include "audiosys.h"
#include "nes6502.h"
#include "m_nsf.h"
#include "nsf6502.h"

#include "sct_main.h"

#define SHIFT_CPS 16
#define NES_BASECYCLES (21477270)
static Uint32 nsf6502_cleft;	/* fixed point */
static Uint32 nsf6502_cps;		/* cycles per sample:fixed point */
static Uint32 nsf6502_cycles;
static Uint32 nsf6502_cpf;		/* cycles per frame */
static Uint8  nsf6502_breaked;

static Uint GetWordLE(Uint8 *p)
{
	return p[0] | (p[1] << 8);
}

static void NSF6502PlaySetup(void)
{
	if (nsf6502_breaked)
	{
		NES6502_CONTEXT c6502;
		Uint8 *nsfhead = NSFGetHeader();
		nsf6502_breaked = 0;
		NES6502BreakPoint(0x4103);
		NSFRomInit(GetWordLE(nsfhead + 0xC));	/* PLAY */
		NES6502ContextGet(&c6502);
		c6502.PC = 0x4100;
		c6502.A = 0x00;
		c6502.S = 0xFF;
		c6502.P = 0x26							/* IRZ */;
		NES6502ContextSet(&c6502);
	}
}

#define EXEC_FRAME 0
#if EXEC_FRAME
static Int32 __fastcall Execute6502(void)
{
	Uint32 cycles;
	nsf6502_cleft += nsf6502_cps;
	cycles = nsf6502_cleft >> SHIFT_CPS;
	nsf6502_cleft &= (1 << SHIFT_CPS) - 1;
	nsf6502_cycles += cycles;
	if (nsf6502_cycles > nsf6502_cpf)
	{
		nsf6502_cycles -= nsf6502_cpf;
		NSF6502PlaySetup();
		nsf6502_breaked = NES6502Execute(nsf6502_cpf);
	}
	return 0;
}
#else
static Int32 __fastcall Execute6502(void)
{
	Uint32 cycles;
	nsf6502_cleft += nsf6502_cps;
	cycles = nsf6502_cleft >> SHIFT_CPS;
	if (!nsf6502_breaked)
	{
		nsf6502_breaked = NES6502Execute(cycles);
	}
	nsf6502_cleft &= (1 << SHIFT_CPS) - 1;
	nsf6502_cycles += cycles;
	if (nsf6502_cycles > nsf6502_cpf)
	{
		nsf6502_cycles -= nsf6502_cpf;
		NSF6502PlaySetup();
	}
	return 0;
}
#endif

static NES_AUDIO_HANDLER nsf6502_audio_handler[] = {
	{ 0, Execute6502, },
	{ 0, 0, },
};

#include "sct_init.h"

static Uint32 DivFix(Uint32 p1, Uint32 p2, Uint32 fix)
{
	Uint32 ret;
	ret = p1 / p2;
	p1  = p1 % p2;/* p1 = p1 - p2 * ret; */
	while (fix--)
	{
		p1 += p1;
		ret += ret;
		if (p1 >= p2)
		{
			p1 -= p2;
			ret++;
		}
	}
	return ret;
}

static void __fastcall NSF6502Reset(void)
{
	NES6502_CONTEXT c6502;
	Uint8 *nsfhead = NSFGetHeader();
	Uint freq = NESAudioFrequencyGet();
	Uint speed = 0;

	if (0 == (nsfhead[0x7a] & 1) && GetWordLE(nsfhead + 0x6E))
		speed = GetWordLE(nsfhead + 0x6E);	/* NTSC tune */
	else if ((nsfhead[0x7a] & 1) && GetWordLE(nsfhead + 0x78))
		speed = GetWordLE(nsfhead + 0x78);	/* PAL  tune */

	if (speed == 0)
		speed = (nsfhead[0x7a] & 1) ? (1000000 / 50) : 0x411A;

	nsf6502_cleft = 0;
	nsf6502_cps = DivFix(NES_BASECYCLES, 12 * freq, SHIFT_CPS);

	nsf6502_cycles = 0;
	nsf6502_cpf = speed * ((Uint)(((1 << 15) * (double)NES_BASECYCLES) / (12.0 * 1000000.0)));
	nsf6502_cpf >>= 15;

	NES6502Reset();
	NES6502BreakPoint(0x4103);
	NSFRomInit(GetWordLE(nsfhead + 0xA));	/* INIT */
	NES6502ContextGet(&c6502);
	c6502.PC = 0x4100;
	c6502.A = NSFGetSongNo() - 1;
	c6502.X = (nsfhead[0x7a] & 1) ? 1 : 0;
	c6502.Y = 0;
	c6502.S = 0xFF;
	c6502.P = 0x26;							/* IRZ */
	NES6502ContextSet(&c6502);
	do {
		nsf6502_breaked = NES6502Execute(~0);
	} while(!nsf6502_breaked);				/* infinite */
	NSF6502PlaySetup();
#if EXEC_FRAME
	nsf6502_breaked = NES6502Execute(nsf6502_cpf);
#endif
}

static NES_RESET_HANDLER nsf6502_reset_handler[] = {
	{ NES_RESET_SYS_LAST, NSF6502Reset, },
	{ 0,                  0, },
};

Uint NSF6502Install(void)
{
	NESAudioHandlerInstall(nsf6502_audio_handler);
	NESResetHandlerInstall(nsf6502_reset_handler);
	return NESERR_NOERROR;
}
