#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <xmms/configfile.h>


#include "nsf.h"

#include "callbacks.h"
#include "interface.h"
#include "support.h"

//utilites function
static gboolean inside_sensitive_widgets(gint x, gint y);

void 
nsfwin_motion_event(GtkWidget *widget, GdkEventMotion *event, gpointer callback_data)
{
	
        XEvent ev;
        gint i = 0;

        XSync(GDK_DISPLAY(), False);

        while (XCheckTypedEvent(GDK_DISPLAY(), MotionNotify, &ev))
        {
                event->x = ev.xmotion.x;
                event->y = ev.xmotion.y;
                i++;
        }
        
	if (dock_is_moving(nsf_win->window))
        {
                dock_move_motion(nsf_win->window, event);
        }else{
                handle_motion_cb(nsf_win->wlist, widget, event);
                draw_nsf_window(FALSE);
        }
        gdk_flush();
	
}



void
nsfwin_release_event(GtkWidget * widget, GdkEventButton *event, gpointer callback_data)
{
	gdk_pointer_ungrab(GDK_CURRENT_TIME);

	/*
	 * The gdk_flush() is just for making sure that the pointer really
	 * gets ungrabbed before calling any button callbacks
	 *
	 */
	
	gdk_flush();

	if (dock_is_moving(nsf_win->window))
	{
		dock_move_release(nsf_win->window);
	}
	handle_release_cb(nsf_win->wlist, widget, event);
	draw_nsf_window(FALSE);
	
}

void
nsfwin_press_event(GtkWidget * widget, GdkEventButton *event, gpointer callback_data)
{
	
	gint mx, my;
	gboolean grab = TRUE;

	mx = event->x;
	my = event->y;

	// Handle mouse wheel events 
	if ((event->button == 4 || event->button == 5) && event->type == GDK_BUTTON_PRESS)
	{
		;
	}

	if (event->button == 1 && event->type == GDK_BUTTON_PRESS && 
	    !inside_sensitive_widgets(event->x, event->y) && 
	    (nsf_cfg.easy_move || event->y < 14))
	{
		gdk_window_raise(nsf_win->window->window);
		//equalizerwin_raise();
		//playlistwin_raise();
		dock_move_press(nsf_win->dock_window_list, nsf_win->window,
		event,TRUE);
	} else if (event->button == 1 && event->type == GDK_2BUTTON_PRESS && 
		   event->y < 14){
                
		//mainwin_set_shade(!cfg.player_shaded);
                if(dock_is_moving(nsf_win->window))
                        dock_move_release(nsf_win->window);
        }else{
		handle_press_cb(nsf_win->wlist, widget, event);
		draw_nsf_window(FALSE);
	}
	
	if (grab){
		gdk_pointer_grab(nsf_win->window->window, 
				 FALSE, 
				 GDK_BUTTON_MOTION_MASK | 
				 GDK_BUTTON_RELEASE_MASK, 
				 GDK_NONE, GDK_NONE, 
				 GDK_CURRENT_TIME);
	}
	
}

void 
nsfwin_focus_in(GtkWidget * widget, GdkEvent * event, gpointer callback_data)
{
        nsf_win->focus = 1;
	/*
        if (!cfg.player_visible)
        {
                if (cfg.playlist_visible)
                        playlistwin_raise();
                else if (cfg.equalizer_visible)
                        equalizerwin_raise();
	}*/
        nsf_win->menubtn->pb_allow_draw = TRUE;
        //nsf_win->minimize->pb_allow_draw = TRUE;
        nsf_win->shade->pb_allow_draw = TRUE;
        nsf_win->close->pb_allow_draw = TRUE;
        draw_nsf_window(TRUE);
}

void
nsfwin_focus_out(GtkWidget * widget, GdkEventButton * event, 
		 gpointer callback_data)
{
        nsf_win->focus = 0;
        nsf_win->menubtn->pb_allow_draw = FALSE;
        //nsf_win->minimize->pb_allow_draw = FALSE;
        nsf_win->shade->pb_allow_draw = FALSE;
        nsf_win->close->pb_allow_draw = FALSE;
        draw_nsf_window(TRUE);
}

void
nsfwin_prev_cb(void)
{
	if(nsf_file){
		if(nsf_file->going){	
			nsf_file->eok = 1;
			nsf_file->eos = 1;
			nsf_file->current_song -= 2;
		}
	}
	
}

void
nsfwin_next_cb(void)
{
	if(nsf_file){
		if(nsf_file->going){	
			nsf_file->eok = 1;
			nsf_file->eos = 1;
		}
	}
}

void
nsfwin_menubtn_cb(void)
{
	nsf_configure();
}

void
nsfwin_minimize_cb(void)
{

	
	Window xwindow;

        if (!nsf_win->window->window){
                return;
	}

        xwindow = GDK_WINDOW_XWINDOW(nsf_win->window->window);
        XIconifyWindow(GDK_DISPLAY(), xwindow, DefaultScreen(GDK_DISPLAY()));
	
}

void 
nsfwin_shade_cb(void)
{
}

void 
nsfwin_close_cb(void)
{
	gtk_widget_hide(nsf_win->window);
}

void
nsf_on_ok_pressed                          (GtkButton       *button,
					      gpointer         user_data)
{
	ConfigFile *cfg;
	gchar *filename;
	GtkWidget *widget;
	
	widget = lookup_widget(GTK_WIDGET(button), "spinbutton1");
	nsf_cfg.play_time =  gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
	widget = lookup_widget(GTK_WIDGET(button), "spinbutton2");
	nsf_cfg.fadeout_time =  gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
	
	filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
	cfg = xmms_cfg_open_file(filename);
	if (!cfg){
		cfg = xmms_cfg_new();
	}
	
	xmms_cfg_write_int(cfg, "NSF", "play_time",nsf_cfg.play_time);
	xmms_cfg_write_int(cfg, "NSF", "fadeout_time",nsf_cfg.fadeout_time);
	xmms_cfg_write_file(cfg, filename);
	xmms_cfg_free(cfg);
	g_free(filename);
	
	widget = lookup_widget(GTK_WIDGET(button), "configure");
	gtk_widget_hide(GTK_WIDGET(widget));
}

gboolean
nsfwin_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer data)
{

	if(nsf_file){
		if(nsf_file->going){	
			nsf_file->eok = 1;
			switch(event->keyval){
			case GDK_a:
				nsf_file->eos = 1;
				nsf_file->current_song -= 2;
				break;
			case GDK_s:
				nsf_file->eos = 1;
				break;
			}
		}
	}
	return TRUE;
}

static gboolean inside_sensitive_widgets(gint x, gint y)
{
	return (inside_widget(x, y, nsf_win->menubtn) || 
		//inside_widget(x, y, nsf_win->minimize) ||
		inside_widget(x, y, nsf_win->shade) ||
		inside_widget(x, y, nsf_win->close) ||
		inside_widget(x, y, nsf_win->rew) || 
		inside_widget(x, y, nsf_win->fwd));
		//inside_widget(x, y, nsf_win->play) ||
		//inside_widget(x, y, nsf_win->pause) || 
		//inside_widget(x, y, nsf_win->stop) ||
		//inside_widget(x, y, nsf_win->eject) ||
		
		
}
