/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* ::::::CAUSION::::
 *  .nsf format
 *  only for 44100Hz, 16bit and mono
 */

#define VERSION  "0.0.3"
#define PLAYTIME 30
#define FADEOUTTIME 5

#include "nsf.h"



#include "interface.h"
#include "callbacks.h"
#include "support.h"

static void nsf_init(void);
static void nsf_about(void);
static int is_our_file(char *filename);
static void play_file(char *filename);
static void stop(void);
static void seek(int time);
static void nsf_pause(short p);
static int get_time(void);
static void get_song_info(char *filename, char **title, int *length);
static void nsf_change_song(gint song_number);
static void draw_nsfwin_titlebar(int focus);

InputPlugin nsf_ip =
{
	NULL,
	NULL,
	"NSF Player " VERSION,
	nsf_init,
	nsf_about,
	nsf_configure,
	is_our_file,
	NULL,
	play_file,
	stop,
	nsf_pause,
	seek,
	NULL,
	get_time,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	get_song_info,
	NULL,			/* file_info_box */
	NULL
};

NsfFile *nsf_file = NULL;
NsfConfig nsf_cfg;
NsfWin nsfwin, *nsf_win = NULL;

static pthread_t decode_thread;
static gboolean audio_error = FALSE;
static unsigned char *buffer = NULL;
static GtkWidget *nsf_configure_win = NULL;

static void
draw_nsfwin_titlebar(int focus)
{
	if (focus){
		gdk_draw_pixmap(nsf_win->bg, nsf_win->gc,
				get_skin_pixmap(SKIN_TITLEBAR), 
				6+27, 3, 
				6, 3, 9, 9 );
		gdk_draw_pixmap(nsf_win->bg, nsf_win->gc,
				get_skin_pixmap(SKIN_TITLEBAR), 
				254+27, 3, 
				254, 3, 9, 9 );
		gdk_draw_pixmap(nsf_win->bg, nsf_win->gc,
				get_skin_pixmap(SKIN_TITLEBAR), 
				264+27, 3, 
				264, 3, 9, 9 );
	}else{
		gdk_draw_pixmap(nsf_win->bg, nsf_win->gc,
				get_skin_pixmap(SKIN_TITLEBAR), 
				6+27, 3+15, 
				6, 3, 9, 9 );
		gdk_draw_pixmap(nsf_win->bg, nsf_win->gc,
				get_skin_pixmap(SKIN_TITLEBAR), 
				254+27, 3+15, 
				254, 3, 9, 9 );
		gdk_draw_pixmap(nsf_win->bg, nsf_win->gc,
				get_skin_pixmap(SKIN_TITLEBAR), 
				264+27, 3+15, 
				264, 3, 9, 9 );
	}
}

void
draw_nsf_window(gboolean force)
{
	GList *wl;
	Widget *w;
	gboolean redraw;
	
	/*
	  if (!cfg.nsfwin_visible)
	  return;
	*/

	lock_widget_list(nsf_win->wlist);
	
	
	if (force){
		/*
		gdk_draw_pixmap(nsf_win->bg, nsf_win->gc, 
				get_skin_pixmap(SKIN_MAIN), 
				0, 0, 0, 0, 275, 
				nsf_cfg.player_shaded ? 14 : 116);
		*/
		gdk_draw_pixmap(nsf_win->bg, nsf_win->gc, 
				get_skin_pixmap(SKIN_MAIN), 
				0, 80, 0, 14, 275, 
				nsf_cfg.player_shaded ? 14 : 36);				
		

		
		gdk_draw_pixmap(nsf_win->bg, nsf_win->gc, 
				get_skin_pixmap(SKIN_EQMAIN), 
				0, 0, 0, 0, 275, 
				nsf_cfg.player_shaded ? 14 : 14);
		
		draw_nsfwin_titlebar(nsf_win->focus);
		draw_widget_list(nsf_win->wlist, &redraw, TRUE);
	}else{
		draw_widget_list(nsf_win->wlist, &redraw, FALSE);
	}

	
	if (redraw || force)
	{

		wl = nsf_win->wlist;
		while (wl)
		{
			w = (Widget *) wl->data;
			if (w->redraw && w->visible)
			{
				gdk_window_clear_area(nsf_win->window->window, 
						      w->x, w->y, 
						      w->width, w->height);
				w->redraw = FALSE;

			}
			wl = wl->next;
		}
	}

	if (force){
		gdk_window_clear(nsf_win->window->window);
	}
	gdk_flush();
	unlock_widget_list(nsf_win->wlist);
}

void
nsf_about(void)
{
	if (!nsf_win->window){
		nsf_win->window = create_about();
		gtk_widget_set_usize (nsf_win->window, 275, 50);
		gtk_widget_set_app_paintable(nsf_win->window, TRUE);	    
		/*
		gtk_window_set_transient_for(GTK_WINDOW(nsf_win->window), 
					     GTK_WINDOW(mainwin));
		*/
		
		
		nsf_win->dock_window_list = 
			dock_add_window(nsf_win->dock_window_list, 
					nsf_win->window);
		/*
		  gtk_window_set_wmclass(GTK_WINDOW(nsf_win->window), 
					"NSF_Controler", "xmms");
		*/
		
		gtk_widget_realize(nsf_win->window);
		gdk_window_set_decorations(nsf_win->window->window, 0);
		nsf_win->gc = gdk_gc_new(nsf_win->window->window);
		
		if (!nsf_win->bg){
			nsf_win->bg = 
				gdk_pixmap_new(nsf_win->window->window, 
					       275, 50, 
					       gdk_rgb_get_visual()->depth);
		
			gdk_window_set_back_pixmap(nsf_win->window->window, 
						   nsf_win->bg, 0);
			nsf_win->menubtn = 
				create_pbutton(&nsf_win->wlist, nsf_win->bg, 
					       nsf_win->gc, 
					       6, 3, 9, 9, 
					       0, 0, 0, 9, 
					       nsfwin_menubtn_cb, 
					       SKIN_TITLEBAR);
			nsf_win->menubtn->pb_allow_draw = FALSE;
        
			
			nsf_win->shade = 
				create_pbutton(&nsf_win->wlist, nsf_win->bg, 
					       nsf_win->gc, 
					       254, 3, 9, 9, 
					       0, 
					       nsf_cfg.player_shaded ? 27 : 18,
					       9, 
					       nsf_cfg.player_shaded ? 27 : 18,
					       nsfwin_shade_cb, 
					       SKIN_TITLEBAR);
			
			nsf_win->shade->pb_allow_draw = FALSE;
			
			nsf_win->close = 
				create_pbutton(&nsf_win->wlist, nsf_win->bg, 
					       nsf_win->gc, 
					       264, 3, 9, 9, 
					       18, 0, 18, 9, 
					       nsfwin_close_cb, 
					       SKIN_TITLEBAR);
			nsf_win->close->pb_allow_draw = FALSE;
			
			nsf_win->rew = 
				create_pbutton(&nsf_win->wlist, nsf_win->bg, 
					       nsf_win->gc, 
					       16, 22, 23, 18, 
					       0, 0, 0, 18, 
					       nsfwin_prev_cb, 
					       SKIN_CBUTTONS);
			nsf_win->fwd = 
				create_pbutton(&nsf_win->wlist, 
					       nsf_win->bg, nsf_win->gc, 
					       39, 22, 22, 18, 
					       92, 0, 92, 18, 
					       nsfwin_next_cb, 
					       SKIN_CBUTTONS);
	
			/*
			nsf_win->rew = 
				create_pbutton(&nsf_win->wlist, nsf_win->bg, 
					       nsf_win->gc, 
					       16, 88, 23, 18, 
					       0, 0, 0, 18, 
					       nsfwin_prev_cb, 
					       SKIN_CBUTTONS);

			nsfwin_play = 
				create_pbutton(&nsfwin_wlist, 
					       nsfwin_bg, nsfwin_gc, 
					       39, 88, 23, 18, 
					       23, 0, 23, 18,
					       nsfwin_play_pushed,
					       SKIN_CBUTTONS);
        
			nsfwin_pause =
				create_pbutton(&nsfwin_wlist,
					       nsfwin_bg, nsfwin_gc, 
					       62, 88, 23, 18, 
					       46, 0, 46, 18, 
					       input_pause, 
					       SKIN_CBUTTONS);
        
			nsfwin_stop = 
				create_pbutton(&nsfwin_wlist, 
					       nsfwin_bg, nsfwin_gc, 
					       85, 88, 23, 18, 
					       69, 0, 69, 18, 
					       nsfwin_stop_pushed, 
					       SKIN_CBUTTONS);
			nsf_win->fwd = 
				create_pbutton(&nsf_win->wlist, 
					       nsf_win->bg, nsf_win->gc, 
					       108, 88, 22, 18, 
					       92, 0, 92, 18, 
					       nsfwin_next_cb, 
					       SKIN_CBUTTONS);

			nsfwin_eject = 
				create_pbutton(&nsfwin_wlist, 
					       nsfwin_bg, nsfwin_gc, 
					       136, 89, 22, 16, 
					       114, 0, 114, 16, 
					       nsfwin_eject_pushed, 
					       SKIN_CBUTTONS);
			*/
		}
		gtk_widget_show(nsf_win->window);
		draw_nsf_window(TRUE);
	}else{
		gtk_widget_show(nsf_win->window);
	}
}

void
nsf_configure()
{
	GtkWidget *widget;
	if (!nsf_configure_win)
	{
		nsf_configure_win = create_configure();

		widget = lookup_widget(nsf_configure_win, "spinbutton1");
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), nsf_cfg.play_time);
		widget = lookup_widget(nsf_configure_win, "spinbutton2");
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), nsf_cfg.fadeout_time);

		gtk_widget_show(nsf_configure_win);
	}else{
		gtk_widget_show(nsf_configure_win);
		gdk_window_raise(nsf_configure_win->window);
	}

}

InputPlugin*
get_iplugin_info(void)
{
	return &nsf_ip;
}

static void 
nsf_init(void)
{
	ConfigFile *cfg;
	gchar *filename;
	
	nsf_win = &nsfwin;
	nsf_win->window = NULL;
	nsf_win->bg = NULL;
	//nsf_win->dock_window_list = dock_window_list;
	nsf_win->dock_window_list = NULL;
	
	nsf_cfg.play_time = PLAYTIME;	         /* second */
	nsf_cfg.fadeout_time = FADEOUTTIME;      /* second */
	nsf_cfg.player_shaded = FALSE;
	nsf_cfg.easy_move = FALSE;
	
	 filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
	
	if ((cfg = xmms_cfg_open_file(filename)) != 0){
		xmms_cfg_read_int(cfg, "NSF", "play_time", &nsf_cfg.play_time);
		xmms_cfg_read_int(cfg, "NSF", "fadeout_time", &nsf_cfg.fadeout_time);
		xmms_cfg_free(cfg);
	}
}


static int read_n_bytes(FILE * file, guint8 * buf, int n)
{

	if (fread(buf, 1, n, file) != n)
	{
		return FALSE;
	}
	return TRUE;
}

static guint32 convert_to_header(guint8 * buf)
{

	return (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
}

static guint16
read_nsf_id(char *filename)
{
	FILE *file;
	guint8 buf[4];
	guint32 head;

	if (!(file = fopen(filename, "rb"))){		
		return FALSE;
	}
	
	if (!read_n_bytes(file, buf, 4)){
		fclose(file);
		return FALSE;
	}
	
	head = convert_to_header(buf);
	if (head != ('N' << 24) + ('E' << 16) + ('S' << 8) + 'M'){
		fclose(file);
		return FALSE;
	}
	read_n_bytes(file,buf,1);
	if(*buf != 0x1A){
		fclose(file);
		return FALSE;
	}
	fclose(file);
	return TRUE;
}

static int
is_our_file(char *filename)
{
	gchar *ext;

	ext = strrchr(filename, '.');
	
	if (ext){
		if (!strcasecmp(ext, ".nsf")){
		  if (read_nsf_id(filename) == TRUE){
				return TRUE;
		  }
		}
	}
	return FALSE;
}

static void *
play_loop(void *arg)
{
	char data[2048 * 2];
	int bytes, blk_size, rate;
	
	//setting NSF
	NSFSetSongNo(nsf_file->current_song);
	NESAudioFrequencySet(44100);
	NESReset();
	
	//16bit/8bit * 1channel * 512 sample = 1024byte
	blk_size = 512 * (nsf_file->bits_per_sample / 8) * nsf_file->channels;
	//44100 * 2channel * 16bit/8bit = 176kbyte
	rate = nsf_file->samples_per_sec * nsf_file->channels * (nsf_file->bits_per_sample / 8);
	
	while (nsf_file->going){
		if (!nsf_file->eof){
			if(!nsf_file->eos){
				bytes = blk_size;
				if(nsf_file->length - nsf_file->position < bytes){
					bytes = nsf_file->length - nsf_file->position;
					nsf_file->eos = 1;
				}
				if (bytes > 0){
					//read 1024byte = 512 sample * 2byte
					NESAudioRender(data, bytes/2);
					nsf_ip.add_vis_pcm(nsf_ip.output->written_time(), (nsf_file->bits_per_sample == 16) ? FMT_S16_LE : FMT_U8,
							   nsf_file->channels, bytes, data);
					
					while (nsf_ip.output->buffer_free() < bytes && nsf_file->going && nsf_file->seek_to == -1)
					  xmms_usleep(10000);
					if (nsf_file->going && nsf_file->seek_to == -1){
						nsf_ip.output->write_audio(data, bytes);
					}
					nsf_file->position += bytes;
				}
			}else{
				if(nsf_file->eok == 0){
					while(nsf_ip.output->buffer_playing()){
                				xmms_usleep(10000);
        				}
				}
				nsf_file->eok = 0;
				nsf_change_song(++nsf_file->current_song);
			}
			
		}else{
			xmms_usleep(10000);
		}
		
		if (nsf_file->seek_to != -1){
			/*
			nsf_file->position = nsf_file->seek_to * rate;
			fseek(nsf_file->file, nsf_file->position + nsf_file->data_offset, SEEK_SET);
			nsf_ip.output->flush(nsf_file->seek_to * 1000);
			*/
			nsf_file->seek_to = -1;
		}

	}	
	pthread_exit(NULL);
}

static void
play_file(char *filename)
{
	int rate,count,i;
	char *temp,*name;
	GString *string;
	
	audio_error = FALSE;

	nsf_file = g_malloc0(sizeof(NsfFile));
	
	if (!(nsf_file->file = fopen(filename, "rb"))){
		return;
	}
	fseek(nsf_file->file, 0, SEEK_END);
	count = ftell(nsf_file->file);
	fseek(nsf_file->file, 0, SEEK_SET);
	
	(void*)buffer = g_malloc0(count);
	for(i=0;i<count;i++){
		*(buffer+i) = fgetc(nsf_file->file);
	}
	
	fseek(nsf_file->file, 0x06, SEEK_SET);
	read_n_bytes(nsf_file->file, &nsf_file->total_songs, 1);
	read_n_bytes(nsf_file->file, &nsf_file->start_song, 1);
	fseek(nsf_file->file, 0x0e, SEEK_SET);
	read_n_bytes(nsf_file->file, nsf_file->name, 32);
	read_n_bytes(nsf_file->file, nsf_file->artist, 32);
	read_n_bytes(nsf_file->file, nsf_file->copyright, 32);
	fclose(nsf_file->file);
	
	if(NSFLoad(buffer, count) != 0){
		//error;
		return;
	}
	
	nsf_file->channels = 1;  
	nsf_file->samples_per_sec = 44100;
	nsf_file->bits_per_sample = 16;
	nsf_file->data_offset = 0;
	nsf_file->position = 0;
	nsf_file->going = 1;
	nsf_file->eok = 0;
	nsf_file->current_song = 1;
	nsf_file->length = (nsf_file->bits_per_sample / 8)
	                    * nsf_file->channels
	                    * nsf_file->samples_per_sec
	                    * nsf_cfg.play_time;
	
	

	if (nsf_ip.output->open_audio((nsf_file->bits_per_sample == 16) ? FMT_S16_LE : FMT_U8, nsf_file->samples_per_sec, nsf_file->channels) == 0){
		audio_error = TRUE;
		fclose(nsf_file->file);
		g_free(nsf_file);
		nsf_file = NULL;
		return;
	}
	if(strlen(nsf_file->name) != 0){
		string = g_string_new(nsf_file->name);
		g_string_sprintfa(string, " (%d/%d)", nsf_file->current_song, nsf_file->total_songs);
		name = g_strdup(string->str);
		g_string_free(string, TRUE);
	}else{
		temp = strrchr(filename, '/');
		if (!temp){
			temp = filename;
		}else{
			temp++;
		}
	
		name = g_malloc(strlen(temp) + 1);
		strcpy(name, temp);
		*strrchr(name, '.') = '\0';
	}
	
	rate = nsf_file->samples_per_sec * nsf_file->channels * (nsf_file->bits_per_sample / 8);
	nsf_ip.set_info(name, -1, 8 * rate, nsf_file->samples_per_sec, nsf_file->channels);
	g_free(name);
	nsf_file->seek_to = -1;
	pthread_create(&decode_thread, NULL, play_loop, NULL);
}


static void
stop(void)
{
	if (nsf_file)
	{
		if (nsf_file->going)
		{
			nsf_file->going = 0;
			pthread_join(decode_thread, NULL);
			NESTerminate();
			nsf_ip.output->close_audio();
			g_free(nsf_file);
			nsf_file = NULL;

		}
	}
}

static void
nsf_pause(short p)
{
	nsf_ip.output->pause(p);
}


static void
seek(int time)
{
	return;
}


static int
get_time(void)
{
	if (audio_error)
		return -2;
	if (!nsf_file)
		return -1;
	/*if (nsf_file->eos && !nsf_ip.output->buffer_playing())
	        return 0;*/
	if (!nsf_file->going || (nsf_file->eof && !nsf_ip.output->buffer_playing()))
		return -1;
	else
	{
		return nsf_ip.output->output_time();
	}
}

static void
get_song_info(char *filename, char **title, int *length)
{
	NsfFile *temp_nsf_file;
	int rate;
	char *temp,*name;
	GString *string;
	
	temp_nsf_file = g_malloc0(sizeof(NsfFile));
	
	if (!(temp_nsf_file->file = fopen(filename, "rb"))){
		return;
	}
	fseek(temp_nsf_file->file, 0x06, SEEK_SET);
	read_n_bytes(temp_nsf_file->file, &temp_nsf_file->total_songs, 1);
	read_n_bytes(temp_nsf_file->file, &temp_nsf_file->start_song, 1);
	
	fseek(temp_nsf_file->file, 0x0e, SEEK_SET);
	read_n_bytes(temp_nsf_file->file, temp_nsf_file->name, 32);
	
	temp_nsf_file->length = -1;
	temp_nsf_file->channels = 1;
	temp_nsf_file->samples_per_sec = 44100;
	temp_nsf_file->bits_per_sample = 16;
	temp_nsf_file->data_offset = 0;
	temp_nsf_file->position = 0;
	temp_nsf_file->going = 1;
	
	rate = temp_nsf_file->samples_per_sec * temp_nsf_file->channels * (temp_nsf_file->bits_per_sample / 8);
	(*length) = -1;
	
	if(strlen(temp_nsf_file->name) != 0){
		string = g_string_new(temp_nsf_file->name);
		g_string_sprintfa(string, " (%d/%d)", temp_nsf_file->start_song, temp_nsf_file->total_songs);
		(*title) = g_strdup(string->str);
		g_string_free(string, TRUE);
	}else{
		temp = strrchr(filename, '/');
		
		if (!temp){
			temp = filename;
		}else{
			temp++;
		}
		name = g_malloc(strlen(temp) + 1);
		strcpy(name, temp);
		*strrchr(name, '.') = '\0';
		(*title) = name;
	}
	fclose(temp_nsf_file->file);
	g_free(temp_nsf_file);
	temp_nsf_file = NULL;
}

static void
nsf_change_song(gint song_number)
{
	if(0 < song_number && song_number <=nsf_file->total_songs){
		gint rate;
		
		NSFSetSongNo(nsf_file->current_song);
		NESReset();
		nsf_ip.output->flush(0);
		nsf_file->position = 0;
		nsf_file->length = 
			(nsf_file->bits_per_sample / 8)
			* nsf_file->channels
			* nsf_file->samples_per_sec
			* nsf_cfg.play_time;
		
		rate = nsf_file->samples_per_sec
			* nsf_file->channels 
			* (nsf_file->bits_per_sample / 8);
		
		if(strlen(nsf_file->name) != 0){
			GString *string;
			string = g_string_new(nsf_file->name);
			g_string_sprintfa(string," (%d/%d)", 
					  nsf_file->current_song, 
					  nsf_file->total_songs);
			
			nsf_ip.set_info(string->str, -1, 8 * rate, 
					nsf_file->samples_per_sec, 
					nsf_file->channels);
			g_string_free(string, TRUE);
		}
	}else{
		nsf_file->eof = 1;
	}
	nsf_file->eos = 0;
}
